# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/i0sc43hrfspfw5qv39q16zbbd77jpp4l-rsyscall"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/i0sc43hrfspfw5qv39q16zbbd77jpp4l-rsyscall",
])
