"""Various thread implementations with various different special abilities.

Note that all thread types can launch any other thread type, in arbitrary combinations.

Note also that users are able to implement their own types of threads,
so if these available thread types do not meet your needs,
you can implement a new kind of thread which does.

"""
