# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/zhbi4fvq3av9h1rc4wd8imzmc2jk0zsv-nix-2.3.10"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/08qdc7hl7gxs1wbk9fynmvrr4zmpaqfl-bzip2-1.0.6.0.2",
    "/nix/store/1pzq6yh1r415ykizk6zw61z5jlkx9994-openssl-1.1.1j",
    "/nix/store/6hhqb94ilrdmh0cx5bhdqb92m6bqkvdj-zlib-1.2.11",
    "/nix/store/85rb8jkj3g9c24k50bjy5vypj45mnq86-nghttp2-1.41.0-lib",
    "/nix/store/w65ydq10abi813fi7d5j7afdcrxj3aqq-bash-4.4-p23",
    "/nix/store/z88sdrj94jhbp7v8vxf1z8xy6vcv8wln-keyutils-1.6.3-lib",
    "/nix/store/f059zv2fl9ml1q53qj5wbwi8xiim28g1-libkrb5-1.18",
    "/nix/store/hhhzkca6bw5d6qh22g2igqx8vfb07a5b-libssh2-1.9.0",
    "/nix/store/kdbkmym34kivwsjkly8d3j5w6rhj2jz9-brotli-1.0.9-lib",
    "/nix/store/1c9xv0m1khm7wwhvg2xa8xkzjwls27gr-curl-7.74.0",
    "/nix/store/2cw85jfh8yj20cj62lcs21y5jfcq28bi-xz-5.2.5",
    "/nix/store/kk5rvl6m9hc3sgmzf7fw5735k6izkbvb-attr-2.4.48",
    "/nix/store/2dx5xnd837631gcdj0z00h7xlv9jp402-acl-2.2.53",
    "/nix/store/ra609mrrl92jirph1zbahcb06pxmjhcl-openssl-1.1.1j-bin",
    "/nix/store/kic4qky3myhflpf8x6zqq332vqj4wa23-openssl-1.1.1j-dev",
    "/nix/store/6kv4p9gnq517qdxhz3cd56fa4waxwgfp-s2n-tls-1.0.0",
    "/nix/store/8pwcw8vwd194sz91iy0lpiv6wizb2xh4-aws-c-common-0.5.2",
    "/nix/store/if0pk349c8c0pflsjad93i93mq489rk7-aws-checksums-0.1.11",
    "/nix/store/wwkm1iy3cwmhjs4a6fnwv2j79iqhc23q-aws-c-cal-0.4.5",
    "/nix/store/smj73q33wc3c3caxvrbmxqlw78i4vdxw-aws-c-io-0.9.1",
    "/nix/store/d37ikmyb8kz18na1vmlqhi85gn93j381-aws-c-event-stream-0.2.7",
    "/nix/store/fmxgxr7kx29aqbzjp662v5mhkrbjvl91-gcc-10.2.0-lib",
    "/nix/store/3r64nnwxn40brxsmvgwrcfj8vdpr65sq-aws-sdk-cpp-1.8.121",
    "/nix/store/89kh3d91b9i9mqh39fj6xdff1sw81m9f-gzip-1.10",
    "/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32",
    "/nix/store/c3vfwsi6ipdgxn4wlfc5qmfjp4d86cvf-nix-2.3.10-man",
    "/nix/store/dajsdxwm5q94rzfv0fkmjbgcshfhm7if-busybox-1.32.1-x86_64-unknown-linux-musl",
    "/nix/store/dppx4515xywkvfkbm2h9iri83h4y5pnq-libsodium-1.0.18",
    "/nix/store/gp9zgl6sra50wjraisc7hywgm62xmxdq-editline-1.17.0",
    "/nix/store/hxpycgfpdnmnc9iwxakxr037k0bvdq97-gnutar-1.32",
    "/nix/store/pqlxb52qjmg4sq2m80zh27ppqr4l0zhx-libseccomp-2.5.1-lib",
    "/nix/store/sg3h64z6a7mxmap8zgrhhjscpri5b5vn-xz-5.2.5-bin",
    "/nix/store/vw1lcig1bkncmsmhif27h8hq1gaxldhg-bzip2-1.0.6.0.2-bin",
    "/nix/store/wblkmksnrpql1g3vlnc0rd44469kj10k-sqlite-3.34.1",
    "/nix/store/wkl16zs8r7mhvc1khyj7ya2hn30x0hxw-boehm-gc-8.0.4",
    "/nix/store/zhbi4fvq3av9h1rc4wd8imzmc2jk0zsv-nix-2.3.10",
])
