# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/kk5rvl6m9hc3sgmzf7fw5735k6izkbvb-attr-2.4.48",
    "/nix/store/2dx5xnd837631gcdj0z00h7xlv9jp402-acl-2.2.53",
    "/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32",
])
