# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/j1hif8jksmdr5200i8c7v4xm24zm262d-openssh-8.4p1"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/08qdc7hl7gxs1wbk9fynmvrr4zmpaqfl-bzip2-1.0.6.0.2",
    "/nix/store/kk5rvl6m9hc3sgmzf7fw5735k6izkbvb-attr-2.4.48",
    "/nix/store/2dx5xnd837631gcdj0z00h7xlv9jp402-acl-2.2.53",
    "/nix/store/w65ydq10abi813fi7d5j7afdcrxj3aqq-bash-4.4-p23",
    "/nix/store/5iqkifw105y41b226dqqb4279isd7bwv-audit-2.8.5",
    "/nix/store/6hx1crbjzqbjmy907p7dzdqq4sbxszp6-libcap-2.47-lib",
    "/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32",
    "/nix/store/6hhqb94ilrdmh0cx5bhdqb92m6bqkvdj-zlib-1.2.11",
    "/nix/store/b6p5cspsidkc2bpljhwz7d0nbmgrx5z0-glibc-2.32-37-bin",
    "/nix/store/fmxgxr7kx29aqbzjp662v5mhkrbjvl91-gcc-10.2.0-lib",
    "/nix/store/j4ai7j4mazjy6lnq1dbmwv98gvjmiq6f-db-4.8.30",
    "/nix/store/im1gjz24adc97l8linz1ryrzg0x1gg1l-linux-pam-1.5.1",
    "/nix/store/9afqpaqknbcqcpiq0vplm6pq9lazk5cc-shadow-4.8.1",
    "/nix/store/bb4l2lmd6vrcyy40ciac8g6wb49a8szp-util-linux-2.36.1",
    "/nix/store/a4qjxgdrsyk8lvb0pwgmjfscbpvifakw-util-linux-2.36.1-bin",
    "/nix/store/2cw85jfh8yj20cj62lcs21y5jfcq28bi-xz-5.2.5",
    "/nix/store/k5s62x4i13ri5v6q75hphahqf1jgl6g0-kmod-27",
    "/nix/store/m9fxvc4gsmprv1cir77m2i942x5jhlb2-kexec-tools-2.0.20",
    "/nix/store/4jllf2xc8cpk25cqmnfkhf3xhfdmgrjk-ncurses-6.2",
    "/nix/store/ycrv6mkzhxl0vj5xvy8hl4ad095hv42s-readline-7.0p5",
    "/nix/store/mshinwyjb1imfpdjmrpbbh3r7zz5sn6x-bash-interactive-4.4-p23",
    "/nix/store/pqlxb52qjmg4sq2m80zh27ppqr4l0zhx-libseccomp-2.5.1-lib",
    "/nix/store/89kh3d91b9i9mqh39fj6xdff1sw81m9f-gzip-1.10",
    "/nix/store/vw1lcig1bkncmsmhif27h8hq1gaxldhg-bzip2-1.0.6.0.2-bin",
    "/nix/store/rrrdjd8qmsravhwdansg969xmss8yc26-kbd-2.0.4",
    "/nix/store/gjxyg641kinpssxyd1nvbrbgd3yi5hg8-systemd-minimal-247.3",
    "/nix/store/1z35cc7x8dg2a1ygjaycqzfaqmd2bf2n-libusb-1.0.24",
    "/nix/store/0m1dvm2dayzz24k7bdjrlq2kgc2ylxrz-hidapi-0.10.1",
    "/nix/store/1pzq6yh1r415ykizk6zw61z5jlkx9994-openssl-1.1.1j",
    "/nix/store/dx296x7d0v2cdbz3ya2mrd67pq8fvd5d-libcbor-0.8.0",
    "/nix/store/1qaww47sckfihaz8z2zds5vcq3zzkjlf-libfido2-1.5.0",
    "/nix/store/z88sdrj94jhbp7v8vxf1z8xy6vcv8wln-keyutils-1.6.3-lib",
    "/nix/store/f059zv2fl9ml1q53qj5wbwi8xiim28g1-libkrb5-1.18",
    "/nix/store/m2hzy9d98gy448f8s47iqq11y37qp3x5-libedit-20210216-3.1",
    "/nix/store/j1hif8jksmdr5200i8c7v4xm24zm262d-openssh-8.4p1",
])
