# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/dx1b537ww44fy4bma5czf05l43wwh9d4-miredo-1.2.6"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/c9v2yc541g7kr4rhd1g8ljdhgadyy7rg-libelf-0.8.13",
    "/nix/store/dfhmwqgdiycwy7a53p77pgxlq6mh0299-libmnl-1.0.4",
    "/nix/store/kj85rj6fnxr5pzpmsn3ddl0m8yyvkjrz-libnfnetlink-1.0.1",
    "/nix/store/17sr4mnf9kii98g40ijs5vw4yqig1sym-libnetfilter_conntrack-1.0.8",
    "/nix/store/w65ydq10abi813fi7d5j7afdcrxj3aqq-bash-4.4-p23",
    "/nix/store/nkyms1w5v9pc99w604hk4d0lmp11zzhh-libpcap-1.10.0",
    "/nix/store/w123mqy21rn5r1ykmrgqdg2aax4g65vh-libnftnl-1.1.9",
    "/nix/store/gjr0vzngij7m2kyk76m6s1g6p7qqx8wa-iptables-1.8.7",
    "/nix/store/fmxgxr7kx29aqbzjp662v5mhkrbjvl91-gcc-10.2.0-lib",
    "/nix/store/z0r840f3b0p4fziwwqm1waf4aq6zj3ib-db-5.3.28",
    "/nix/store/0xk9cxmnsi3zypvnpfqbhjlif8v761nj-iproute2-5.11.0",
    "/nix/store/sxrmj115nllx5sl510lfii53b49gfr08-judy-1.0.5",
    "/nix/store/dx1b537ww44fy4bma5czf05l43wwh9d4-miredo-1.2.6",
])
